import numpy as np
from math import factorial
import fc_hypermesh.CartesianGrid as CG

def KuhnTriangulation(d):
  q=CG.points(np.ones(d))
  qref=np.vstack([np.zeros((1,d)),np.tril(np.ones((d,d)))]).T
  P=CG.perms(np.arange(d))
  nme=factorial(d);nq=2**d
  me=np.zeros((d+1,nme),dtype=int)
  A=np.matrix(2**np.arange(d))
  for k in range(nme):
    me[:,k]=A*qref[P[k]]
    ql=q[:,me[:,k]]
    s=orientation(ql)
    if (s==-1):
      tmp=me[0,k]
      me[0,k]=me[d,k]
      me[d,k]=tmp
  return q,me

def orientation(ql):
  #d=ql.shape[1]
  #D=np.c_[np.ones((d+1,1)),ql]
  d=ql.shape[0]
  D=np.r_[np.ones((1,d+1)),ql]
  return np.sign(np.linalg.det(D))

def Faces(N,m):
  N=np.array(N)
  d=len(N)
  sTh=CG.TriFaces(N,m)
  for k in range(len(sTh)):
    for i in range(d):
      sTh[k].q[i]/=N[i]
  return sTh

def Triangulation(N):
  d=len(N)
  q,me=CG.Triangulation(N)
  for i in range(d):
    q[i,:]/=N[i]
  return q,me


""" demos of the OrthMesh class (see fc_hypermesh.OrthMesh module)

Usage information
=================
There are seven available functions:

    1. demo01: [-1,1]x[-1,1]x[-1,1] meshing by orthotopes
    2. demo02: [-1,1]x[-1,1] meshing by orthotopes
    3. demo03: [-1,1]x[0,1]x[0,2] meshing by orthotopes
    4. demo04: [-1,1]x[0,1] meshing by simplices
    5. demo05: [0,1]x[0,1] mapped by (x,y)->(2*x,y) and meshed by simplices
    6. demo06: [0,1]x[0,1] mapped by (x,y)->(20*x,2*(2*y-1+cos(2*pi*x))) 
       and meshed by simplices
    7. demo07: [0,1]x[0,1]x[0,1] mapped by 
       (x,y,z)->(x+sin(4*pi*y),10*y-1,z+cos(4*pi*y)) 
       and meshed by simplices
    
Example 1
---------
>>>   import fc_hypermesh.demos
>>>   demos.demo03()

"""
from fc_tools.colors import str2rgb
from fc_tools.others import isModuleFound
from fc_hypermesh.OrthMesh import OrthMesh

if isModuleFound('matplotlib'):
  import matplotlib.pyplot as plt
  from fc_tools.Matplotlib import DisplayFigures,set_axes_equal,SaveAllFigsAsFiles

  def savefigs(basename,**kwargs):
    savedir=kwargs.get('savedir', None )
    if savedir is not None:
      SaveAllFigsAsFiles(basename,dir=savedir)
      
Show=isModuleFound('matplotlib')

def alldemos(show=Show,**kwargs):
  ListOfDemos=['demo01','demo02','demo03','demo04','demo05','demo06','demo07']
  for demo in ListOfDemos:
    rundemo(demo,show,**kwargs)
  savedir=kwargs.get('savedir', None )
  if savedir is not None:  
    print('  -> All figures save in %s'%savedir)

def demo01(show=Show,stop=False):
  """ OrthMesh  """
  print('------ demo01 --------')
  print("oTh=OrthMesh(3,[10,5,10],type='orthotope', box=[[-1,1],[0,1],[0,2]])")
  print('----------------------')
  oTh=OrthMesh(3,[10,5,10],type='orthotope', box=[[-1,1],[0,1],[0,2]])
  print(oTh)
  
  if show:
    if not isModuleFound('matplotlib'):
      print('[fc-hypermesh] Needs matplotlib package to be installed for graphics')
      return
  
    plt.close('all')
    plt.ion()
    fig=plt.figure(1)
    oTh.plotmesh(legend=True)
    set_axes_equal()
    plt.figure(2)
    oTh.plotmesh(m=2,legend=True,edgecolor=[0,0,0])
    set_axes_equal()
    plt.axis('off')

    plt.figure(3)
    oTh.plotmesh(m=2,facecolor=None,edgecolor=str2rgb('LightGray'))
    oTh.plotmesh(m=1,legend=True,linewidth=2)
    set_axes_equal()
    plt.axis('off')

    plt.figure(4)
    oTh.plotmesh(m=1,color='black')
    oTh.plotmesh(m=0,legend=True,s=55) # see matplotlib.pyplot.scatter options
    set_axes_equal()
    plt.axis('off')
    #plt.show(block=stop)
  
def demo02(show=Show,stop=False):
  print('------ demo02 --------')
  print("oTh=OrthMesh(2,[12,5],type='orthotope',box=[[-1,1],[0,1]])")
  print('----------------------')
  oTh=OrthMesh(2,[12,5],type='orthotope',box=[[-1,1],[0,1]])
  print(oTh)
  
  if show:
    if not isModuleFound('matplotlib'):
      print('[fc-hypermesh] Needs matplotlib package to be installed for graphics')
      return
  
    plt.close('all')
    plt.ion()
    plt.figure(1)
    oTh.plotmesh(legend=True)
    set_axes_equal()

    plt.figure(2)
    oTh.plotmesh(m=1,legend=True,linewidth=3)
    set_axes_equal()
    plt.axis('off')

    plt.figure(3)
    oTh.plotmesh(m=1,color='black')
    oTh.plotmesh(m=0,legend=True,s=105) # see matplotlib.pyplot.scatter options
    set_axes_equal()
    plt.axis('off')  
    #plt.show(block=stop)
  
def demo03(show=Show,stop=False):
  print('------ demo03 --------')
  print("oTh=OrthMesh(3,[10,5,10],box=[[-1,1],[0,1],[0,2]])")
  print('----------------------')
  oTh=OrthMesh(3,[10,5,10],box=[[-1,1],[0,1],[0,2]])
  print(oTh)
  
  if show:
    if not isModuleFound('matplotlib'):
      print('[fc-hypermesh] Needs matplotlib package to be installed for graphics')
      return
  
    plt.close('all')
    plt.ion()
    plt.figure(1)
    oTh.plotmesh(legend=True,linewidth=0.5)
    set_axes_equal()

    plt.figure(2)
    oTh.plotmesh(m=2,legend=True,edgecolor=[0,0,0])
    set_axes_equal()
    plt.axis('off')

    plt.figure(3)
    oTh.plotmesh(m=2,edgecolor=[0,0,0],color='none')
    oTh.plotmesh(m=1,legend=True,linewidth=2,alpha=0.3)
    set_axes_equal()
    plt.axis('off')

    plt.figure(4)
    oTh.plotmesh(m=1,color='black',alpha=0.3)
    oTh.plotmesh(m=0,legend=True,s=55)
    set_axes_equal()
    plt.axis('off')
    plt.show(block=stop)

def demo04(show=Show,stop=False):
  print('------ demo04 --------')
  print("oTh=OrthMesh(2,[12,5],type='simplicial',box=[[-1,1],[0,1]])")
  print('----------------------')
  oTh=OrthMesh(2,[12,5],type='simplicial',box=[[-1,1],[0,1]])
  print(oTh)
  
  if show:
    if not isModuleFound('matplotlib'):
      print('[fc-hypermesh] Needs matplotlib package to be installed for graphics')
      return
  
    plt.close('all')
    plt.ion()
    plt.figure(1)
    oTh.plotmesh(legend=True)
    set_axes_equal()

    plt.figure(2)
    oTh.plotmesh(m=1,legend=True,linewidth=3)
    plt.axis('off')
    set_axes_equal()

    plt.figure(3)
    oTh.plotmesh(m=1,color='black')
    oTh.plotmesh(m=0,legend=True,s=105) # see matplotlib.pyplot.scatter options
    plt.axis('off')
    set_axes_equal()
    plt.show(block=stop)
    
def demo05(show=Show,stop=False):
  import numpy as np
  print('------ demo05 --------')
  print('trans=lambda q: np.array([2*q[0],q[1]])')
  print("oTh=OrthMesh(2,5,type='simplicial',mapping=trans)")
  print('----------------------')
  trans=lambda q: np.array([2*q[0],q[1]])
  oTh=OrthMesh(2,5,type='simplicial',mapping=trans)
  print(oTh)
  
  if show:
    if not isModuleFound('matplotlib'):
      print('[fc-hypermesh] Needs matplotlib package to be installed for graphics')
      return
  
    plt.close('all')
    plt.ion()
    plt.figure(1)
    oTh.plotmesh(legend=True)
    set_axes_equal()

    plt.figure(2)
    oTh.plotmesh(color='lightgray')
    oTh.plotmesh(m=1,legend=True,linewidth=3)
    plt.axis('equal')
    plt.axis('off')

    plt.figure(3)
    oTh.plotmesh(color='lightgray')
    oTh.plotmesh(m=1,color='black')
    oTh.plotmesh(m=0,legend=True,s=105) # see matplotlib.pyplot.scatter options
    plt.axis('equal')
    plt.axis('off')
    plt.show(block=stop)
      
def demo06(show=Show,stop=False):
  import numpy as np
  print('------ demo06 --------')
  print('trans=lambda q: np.array([20*q[0],2*(2*q[1]-1+np.cos(2*np.pi*q[0]))])')
  print("oTh=OrthMesh(2,[100,20],type='simplicial',mapping=trans")
  print('----------------------')
  trans=lambda q: np.array([20*q[0],2*(2*q[1]-1+np.cos(2*np.pi*q[0]))])
  oTh=OrthMesh(2,[100,20],type='simplicial',mapping=trans)
  print(oTh)

  if show:
    if not isModuleFound('matplotlib'):
      print('[fc-hypermesh] Needs matplotlib package to be installed for graphics')
      return
  
    plt.close('all')
    plt.ion()
    plt.figure(1)
    oTh.plotmesh(legend=True)
    plt.axis('equal')

    plt.figure(2)
    oTh.plotmesh(color='lightgray')
    oTh.plotmesh(m=1,legend=True,linewidth=3)
    plt.axis('equal')
    plt.axis('off')

    plt.figure(3)
    oTh.plotmesh(color='lightgray')
    oTh.plotmesh(m=1,color='black')
    oTh.plotmesh(m=0,legend=True,s=105) # see matplotlib.pyplot.scatter options
    plt.axis('equal')
    plt.axis('off')
    #plt.show(block=stop)

def demo07(show=Show,stop=False):
  import numpy as np
  print('------ demo07 --------')
  print('trans=lambda q: np.array([q[0]+np.sin(4*np.pi*q[1]), 10*q[1]-1, q[2]+np.cos(4*np.pi*q[1])])')
  print("oTh=OrthMesh(3,[3,25,3],type='simplicial',mapping=trans)")
  print('----------------------')
  trans=lambda q: np.array([q[0]+np.sin(4*np.pi*q[1]), 10*q[1]-1, q[2]+np.cos(4*np.pi*q[1])])
  oTh=OrthMesh(3,[3,25,3],type='simplicial',mapping=trans)
  print(oTh)
  
  if show:
    if not isModuleFound('matplotlib'):
      print('[fc-hypermesh] Needs matplotlib package to be installed for graphics')
      return
  
    plt.close('all')
    plt.ion()
    plt.figure(1)
    oTh.plotmesh(legend=True)
    set_axes_equal()

    plt.figure(2)
    oTh.plotmesh(m=2,legend=True,edgecolor=[0,0,0])
    set_axes_equal()

    plt.figure(3)
    oTh.plotmesh(m=2,edgecolor='lightgray',facecolor=None,alpha=0.3)
    oTh.plotmesh(m=1,legend=True,linewidth=2)
    set_axes_equal()

    plt.figure(4)
    oTh.plotmesh(m=1,color='black')
    oTh.plotmesh(m=0,legend=True,s=55)
    set_axes_equal()
    plt.show(block=stop)
    
    
def rundemo(demo,show,**kwargs):
  print('[fc_hypermesh] Running %s'%demo)
  eval(demo+'(show)',globals(),locals())
  if show:
    plt.show()
    seedemo(show)
    savefigs(demo,**kwargs)
    plt.close('all')
  
def seedemo(show):
  if show:
    DisplayFigures()
    plt.show()
    #import time
    print('    Waiting 3s before closing ...')
    plt.pause(3)
    #time.sleep(3)
  



""" =======
    EltMesh
    =======
    
    Only contains the `EltMesh` class used by `OrthMesh` class.
"""

from fc_tools.others import isModuleFound

class EltMesh:
  """ class EltMesh
  
      This class is used to store an elementary mesh given by its vertices array `q` and 
      its connectivity array `me`
      
      see description in `Class object OrthMesh` in the report.
  """
  def __init__(self,d,m,q,me,**kwargs):
    import numpy as np
    color=kwargs.get('color', [0,0,1] )
    label=kwargs.get('label', '' )
    type=kwargs.get('type', None)
    toGlobal=kwargs.get('toGlobal', None)
    assert  m <= d 
    assert q.shape[0]==d 
    self.d=d
    self.m=m
    self.q=q
    self.me=me
    if type is None:
      if (me.shape[0]==m+1): # m-simplicial
        self.type=0
      elif (me.shape[0]==2**m): # m-orthotope
        self.type=1;
      else:
        raise NameError('Trouble with "me" dimension!')
    else:
      assert type in [0,1]
      assert ( (type==0) and (me.shape[0]==m+1) ) or ( (type==1) and (me.shape[0]==2**m) ) 
      self.type=type
      
    self.nq=q.shape[1]
    self.nme=me.shape[1]
    if toGlobal is None:
      self.toGlobal=None
    else:
      self.toGlobal=np.array(toGlobal,dtype=int)
    
    self.color=color
    self.label=label
  def __repr__(self):
    strret = ' %s object \n'%self.__class__.__name__ 
    strret += '    type (str): %s\n'%self.strtype()
    strret += '    type : %d\n'%self.type
    strret += '       d : %d\n'%self.d 
    strret += '       m : %d\n'%self.m
    strret += '       q : (%d,%d)\n'%self.q.shape
    strret += '      me : (%d,%d)\n'%self.me.shape
   # strret += 'toGlobal : (%d,)\n'%self.toGlobal.shape
    return strret  

  def strtype(self):
    if self.type==0:
      return 'simplicial'
    if self.type==1:  
      return 'orthotope'
    return 'unknow'
  
  if isModuleFound('matplotlib'):
    from .Matplotlib import plotmesh




import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Polygon,Patch
from mpl_toolkits.mplot3d import Axes3D
import mpl_toolkits.mplot3d as a3
from mpl_toolkits.mplot3d.art3d import Poly3DCollection,Line3DCollection
from fc_tools.others import LabelBaseName
import fc_hypermesh.CartesianGrid as CG

def plotmesh(self,**kwargs):
    color=kwargs.pop('color',self.color)
    if self.type==0:
      return eval("PlotmeshSubTh"+str(self.m)+"simp"+str(self.d)+"D(self,color,**kwargs)")
    else:
      return eval("PlotmeshSubQh"+str(self.m)+"orth"+str(self.d)+"D(self,color,**kwargs)")

def PlotmeshSubTh2simp2D(self,color,**kwargs):
  from matplotlib.patches import Polygon
  from matplotlib.collections import PolyCollection
  Name=LabelBaseName(2,2)
  Poly2D=self.q[:,self.me].swapaxes(0,2)
  p = PolyCollection(Poly2D, facecolor="none",edgecolor = color)
  ax=plt.gca()
  ax.add_collection(p)
  ax.autoscale()
  legend_handle=Patch(color=color,**kwargs) # to improve
  label=r"$%s_{"%(Name)+str(int(self.label))+"}$"
  return legend_handle,label
  
def PlotmeshSubQh2orth2D(self,color,**kwargs):
  from matplotlib.patches import Polygon
  from matplotlib.collections import PolyCollection
  Name=LabelBaseName(2,2)
  Poly2D=self.q[:,self.me[[0,1,3,2],:]].swapaxes(0,2)
  p = PolyCollection(Poly2D, facecolor="none",edgecolor = color)
  ax=plt.gca()
  legend_handle=ax.add_collection(p)
  ax.autoscale()
  legend_handle=Patch(color=color,**kwargs) # to improve
  label=r"$%s_{"%(Name)+str(int(self.label))+"}$"
  return legend_handle,label
  
def PlotmeshSubTh1simp2D(self,color,**kwargs):
  ive=range(self.m+1)
  ANone=np.array([None]*self.nme)
  Name=LabelBaseName(2,1)
  X=np.array([self.q[0,self.me[0]],self.q[0,self.me[1]],ANone]).T.reshape((self.nme*3,))
  Y=np.array([self.q[1,self.me[0]],self.q[1,self.me[1]],ANone]).T.reshape((self.nme*3,))
  fig = plt.gcf()
  ax = fig.gca()
  #plt.rc('text', usetex=True)
  label=r"$%s_{"%(Name)+str(int(self.label))+"}$"
  legend_handle,=ax.plot(X,Y,color=color,picker=5,**kwargs)
  legend_handle.aname=label
  #legend_handle=plt.Line2D([0, 1],[0,1],color=color,**kwargs)
  #label=r"$%s_{"%(Name)+str(int(self.label))+"}$"
  return legend_handle,label

def PlotmeshSubQh1orth2D(self,color,**kwargs):
  return PlotmeshSubTh1simp2D(self,color,**kwargs)

def PlotmeshSubTh0simp2D(self,color,**kwargs):
  s=kwargs.get('s', 20 );kwargs.pop('s',None)
  marker=kwargs.get('marker', 'o' );kwargs.pop('marker',None)
  Name=LabelBaseName(2,0)
  fig = plt.gcf()
  ax = fig.gca()
  label=r"$%s_{"%(Name)+str(int(self.label))+"}$"
  #legend_handle,=ax.plot(self.q[0,0],self.q[1,0],color=color,linestyle='',marker='o',picker=5,**kwargs)
  legend_handle=ax.scatter(self.q[0,0],self.q[1,0],c=color,marker=marker,s=s,**kwargs)
  legend_handle.aname=label
  return legend_handle,label

def PlotmeshSubQh0orth2D(self,color,**kwargs):
  return PlotmeshSubTh0simp2D(self,color,**kwargs)

def PlotmeshSubTh0simp3D(self,color,**kwargs):
  s=kwargs.get('s', 5 );kwargs.pop('s',None)
  marker=kwargs.get('marker', 'o' );kwargs.pop('marker',None)
  Name=LabelBaseName(3,0)
  fig = plt.gcf()
  if len(fig.axes)>0:
    ax=fig.axes[0]
  else:
    ax = fig.gca( projection='3d')
  label=r"$%s_{"%(Name)+str(int(self.label))+"}$"
  #legend_handle=ax.scatter(self.q[0,0],self.q[1,0],self.q[2,0],c=color,marker=marker,picker=picker,**kwargs)
  legend_handle=ax.scatter(self.q[0,0],self.q[1,0],self.q[2,0],c=color,marker=marker,s=s,**kwargs)
  legend_handle.aname=label
  return legend_handle,label

def PlotmeshSubQh0orth3D(self,color,**kwargs):
  return PlotmeshSubTh0simp3D(self,color,**kwargs)
    
def PlotmeshSubTh1simp3D(self,color,**kwargs):
  Name=LabelBaseName(3,1)
  Line3D=self.q[:,self.me].swapaxes(0,2)
  fig = plt.gcf()
  if len(fig.axes)>0:
    ax=fig.axes[0]
  else:
    ax = fig.gca( projection='3d')
  ax.add_collection3d(Line3DCollection(Line3D,colors=color, **kwargs))
  legend_handle=plt.Line2D([0, 1],[0,1],color=color,**kwargs)
  label=r"$%s_{"%(Name)+str(int(self.label))+"}$"
  return legend_handle,label

def PlotmeshSubQh1orth3D(self,color,**kwargs):
  return PlotmeshSubTh1simp3D(self,color,**kwargs)
  
def PlotmeshSubTh2simp3D(self,color,**kwargs): 
  return PlotmeshSubGen2orth3D(self,color,**kwargs)


def PlotmeshSubQh2orth3D(self,color,**kwargs):  
  return PlotmeshSubGen2orth3D(self,color,**kwargs)

def PlotmeshSubGen2orth3D(self,color,**kwargs):  
  Name=LabelBaseName(3,2)
  edgecolor=kwargs.pop('edgecolor',None)
  if color is None:
    color=self.color
  if 'facecolor' in kwargs:
    color=kwargs.pop('facecolor',None)
  if color is None: 
    Line3D=self.q[:,self.me].swapaxes(0,2)
    if self.type==0:
      C=[[0,1],[1,2],[2,0]]
    else:
      C=[[0,1],[2,3],[0,2],[1,3]]
    Poly3D=np.zeros((0,2,3))
    if edgecolor is None:
      edgecolor=self.color
    for edge in C:
      Poly3D=np.concatenate((Poly3D,self.q[:,self.me[edge]].T))
    p = Line3DCollection(Poly3D,color=edgecolor,**kwargs)  
    legend_handle=Patch(color=edgecolor,**kwargs)
  else:  
    if self.type==0:
      Poly3D=self.q[:,self.me].swapaxes(0,2)
    else:  
      Poly3D=self.q[:,self.me[[0,1,3,2]]].swapaxes(0,2)
    p = Poly3DCollection(Poly3D, facecolor=color,edgecolor = edgecolor,**kwargs)
    legend_handle=Patch(color=color,**kwargs)
  label=r"$%s_{"%(Name)+str(int(self.label))+"}$"
  fig = plt.gcf()
  if len(fig.axes)>0:
    ax=fig.axes[0]
  else:
    ax = fig.gca( projection='3d')
  ax.add_collection3d(p)
  ax.autoscale()
  return legend_handle,label
  
def PlotmeshSubTh3simp3D(self,color,**kwargs):
  Name=LabelBaseName(3,3)
  Poly3D=[]
  C=CG.combs(np.arange(4),2)
  for i in range(6):
     A=self.q[:,self.me[C[i]]].T
     Poly3D+= A.tolist()
      
  fig = plt.gcf()
  if len(fig.axes)>0:
    ax=fig.axes[0]
  else:
    ax = fig.gca( projection='3d')
  
  ax.add_collection3d(Line3DCollection(Poly3D,colors=color,**kwargs))#linewidths=0.4, linestyles=':')) # linewidths=0.4, linestyles=':')
  #legend_handle,=plt.plot([0, 1],[0,1],color=color,visible=False,**kwargs)
  legend_handle=Patch(color=color,**kwargs) # to improve
  label=r"$%s_{"%(Name)+str(int(self.label))+"}$"
  return legend_handle,label

def PlotmeshSubQh3orth3D(self,color,**kwargs):
  Name=LabelBaseName(3,3)
  Poly3D=[]
  C=[[0,1],[0,4],[1,5],[4,5],[2,3],[2,6],[3,7],[6,7],[1,3],[5,7],[4,6],[0,2]]
  Poly3D=np.zeros((0,2,3))
  for edge in C:
     A=self.q[:,self.me[edge]].T
     Poly3D=np.concatenate((Poly3D,A))
  fig = plt.gcf()
  if len(fig.axes)>0:
    ax=fig.axes[0]
  else:
    ax = fig.gca( projection='3d') 
  ax.add_collection3d(Line3DCollection(Poly3D,colors=color,**kwargs))#linewidths=0.4, linestyles=':')) # linewidths=0.4, linestyles=':')
  ax.autoscale()
  #legend_handle,=plt.plot([0, 1],[0,1],color=color,visible=False,**kwargs)
  legend_handle=Patch(color=color,**kwargs) # to improve
  label=r"$%s_{"%(Name)+str(int(self.label))+"}$"
  return legend_handle,label


""" ============================
    Cartesian Grid Tessallations
    ============================
    
    Provides tessellation of a cartesian grid CG(N)=[0,N1]x...x[0,Nd] with unit hypercubes 
    or with simplices.
    
    Explanation can be found in the report 'Vectorized algorithms for regular tessellations of 
    d-orthotopes and their faces' by F. Cuvelier and available at http:\\...

    ==================== =========================================================
    Utility functions
    ==============================================================================
    perms                Generate all permutations of V.
                         See `Perms`function in the report.
    combs                Generate all combinations of V taken k at a time.
                         See `Combs` function in the report.
    betafunc             Compute the 'beta' integer array.
                         See `CGbeta` function in the report.
    
    ==================== =========================================================
    
    ==================== =========================================================
    Tessallation functions
    ==============================================================================
    points               Compute the points of the Cartesian grid CG(N)
                         See `CartesianGridPoints` function in the report.
    TessHyp              Cartesian grid CG(N) tessallation with unit hypercube
                         see `CartesianGridTessHyp` function in the report 
    KuhnTriangulation    Kuhn's triangulation of the unit d-hypercube [0,1]^d with
                         d! simplices. 
                         See `KuhnTriangulation` function in the report.
    Triangulation        Cartesian grid CG(N) tessallation with simplices.
                         see `CartesianGridTessHyp` function in the report
    TessFaces            Computes all m-faces tessellations of the cartesian grid 
                         CG(N) with unit m-hypercubes.
                         see `CGTriFaces` function in the report
    TriFaces             Computes all m-faces tessellations of the cartesian grid 
                         CG(N) with unit m-simplices.
                         see `CGTessFaces` function in the report
    ==================== =========================================================
  
"""

import numpy as np
import itertools
from math import factorial


def perms(V):
  """ P=perms(V)
  Generate all permutations of V, one row per permutation. The lexicographical 
  order is choosen. The result is a numpy array with size N!-by-N, where N 
  is the length of V. 
  
  Parameters
  ----------
  V : list or array of N scalars
  
  Returns
  -------
  P : (N!,N) shaped ndarray
  
  As an example, 'perms ([1, 2, 3])' returns the numpy array
  
  >>> P=perms([1, 2, 3])
  >>> P
  array([[1, 2, 3],
         [1, 3, 2],
         [2, 1, 3],
         [2, 3, 1],
         [3, 1, 2],
         [3, 2, 1]])    
          
  See `Perms` function in the report.
  """
  return np.array([x for x in itertools.permutations(V,len(V))])

def combs(V,k):
  """
  C=combs(V,k)
  
  Generate all combinations of V taken k at a time, one row per combination. 
  The lexicographical order is choosen. The result is a numpy array with 
  size N!/(k!(n-k)!)-by-k, where N is the length of V and 1<=k<=N.
  
  Parameters
  ----------
  V : list or array of N scalars
  k : integer (0<=k<=N)
  
  Returns
  -------
  C : (N!/(k!(n-k)!),k) shaped ndarray
  
  As an example:
  
  >>> C=combs([1,2,3,4],2)
  >>> C  
  array([[1, 2],
         [1, 3],
         [1, 4],
         [2, 3],
         [2, 4],
         [3, 4]])
    
  See `Combs` function in the report.
  """
  assert (len(V)>=k) and (k>=1)
  return np.array([x for x  in itertools.combinations(V,k)])

def betafunc(N):
  """ beta=betafunc(N)
  
      Compute the 'beta' integer array of size `len(N)` such that
        beta[0]=1 
      and for all l in {1,...,len(N)-1} 
        beta[l] = beta[l-1]*(N[l-1]+1)  
      
      See `CGbeta` function in the report.
  """
  N=np.array(N,dtype=int)
  d=len(N)
  beta=np.ones((d,),dtype=int)
  for l in np.arange(1,d):
    beta[l]=beta[l-1]*(N[l-1]+1)
  return beta

def points(N):
  """ q=points(N)
  
      Compute a d-by-nq array which contains all the points of the cartesian grid CG(N)
      with ``d=len(N)`` and ``nq = (N[0]+1)*...*(N[d]+1)``
      
      Parameters
      ----------
      N : list or array of d integers. The cartesian grid CG(N) is then defined on
             [0,N[0]]x...x[O,N[d-1]]
      
      Returns
      -------
      q : vertices array of the triangulation of CG(N), (d,nq)-shaped ndarray of 
          floats where ``nq=prod(N+1)`` and ``d=len(N)``
      
      As an example:
  
      >>> q=points([10,7,9,13])
      >>> q
      array([[  0.,   1.,   2., ...,   8.,   9.,  10.],
            [  0.,   0.,   0., ...,   7.,   7.,   7.],
            [  0.,   0.,   0., ...,   9.,   9.,   9.],
            [  0.,   0.,   0., ...,  13.,  13.,  13.]])      
            
      See `CartesianGridPoints` function in the report.
  """
  N=np.array(N,dtype=int)
  d=len(N)
  beta=betafunc(N)
  q=np.zeros((d,np.prod(N+1))) # Must be float when using with MappingBox function
  for r in range(d):
    A=np.tile(np.arange(N[r]+1),(beta[r],1)).flatten(1)
    q[r,:]=np.tile(A,(1,np.prod(N[r+1:]+1)))
  return q

def TessHyp(N):
  """ q,me=TessHyp(N)
  
      Computes the vertices array ``q`` and the connectivity array ``me`` obtained 
      from a tesselation of the cartesian grid CG(N) with unit hypercube.
      
      Parameters
      ----------
      N : list or array of d integers. The cartesian grid CG(N) is then defined on
             [0,N[0]]x...x[O,N[d-1]]
      
      Returns
      -------
      q : vertices array of the triangulation of CG(N), (d,nq)-shaped ndarray of 
          integers where ``nq=prod(N+1)`` and ``d=len(N)``
      me: connectivity array of the triangulation of CG(N), (d+1,nme) shaped ndarray 
          of integers  where ``nme=prod(N). 
      
      As an example:
  
      >>> q,me=TessHyp([10,7,9,13])
      >>> q
      array([[  0.,   1.,   2., ...,   8.,   9.,  10.],
            [  0.,   0.,   0., ...,   7.,   7.,   7.],
            [  0.,   0.,   0., ...,   9.,   9.,   9.],
            [  0.,   0.,   0., ...,  13.,  13.,  13.]])
      >>> me
      array([[    0,     1,     2, ..., 11337, 11338, 11339],
            [    1,     2,     3, ..., 11338, 11339, 11340],
            [   11,    12,    13, ..., 11348, 11349, 11350],
            ..., 
            [  969,   970,   971, ..., 12306, 12307, 12308],
            [  979,   980,   981, ..., 12316, 12317, 12318],
            [  980,   981,   982, ..., 12317, 12318, 12319]])      
      
      See `CGTessHyp` function in the report.
  """
  N=np.array(N,dtype=int)
  d=len(N)
  q=points(N)
  Hinv=points(N-1)
  qhat=points(np.ones(d))
  beta=betafunc(N)
  iBase=beta.dot(Hinv) # array index start to 0 
  me=np.zeros((2**d,len(iBase)),dtype=np.int)
  for l in range(2**d):
    me[l,:]=iBase+beta.dot(qhat[:,l])
  return q,me

def KuhnTriangulation(d):
  """ q,me=KuhnTriangulation(d)
  
      Kuhn's triangulation of the unit d-hypercube [0,1]^d with d! simplices.
      (positive orientation) 

      Parameters
      ----------
      d : space dimension
      
      Returns
      -------
      q : vertices array of the unit d-hypercube [0,1]^d. 
          (d,2^d)-shaped ndarray of integers.
      me: connectivity array of the unit d-hypercube [0,1]^d. 
          (d+1,d!) shaped ndarray of integers.
  
      As firt example:
      
      >>> q,me=KuhnTriangulation(2)
      >>> q 
      array([[ 0.,  1.,  0.,  1.],
             [ 0.,  0.,  1.,  1.]])
      >>> me
      array([[0, 3],
             [1, 2],
             [3, 0]])
                   
      As second example:
      
      >>> q,me=KuhnTriangulation(3)
      >>> q 
      array([[ 0.,  1.,  0.,  1.,  0.,  1.,  0.,  1.],
             [ 0.,  0.,  1.,  1.,  0.,  0.,  1.,  1.],
             [ 0.,  0.,  0.,  0.,  1.,  1.,  1.,  1.]])
      >>> me
      array([[0, 7, 7, 0, 0, 7],
             [1, 1, 2, 4, 2, 4],
             [3, 5, 3, 5, 6, 6],
             [7, 0, 0, 7, 7, 0]])            
                       
      See `KuhnTriangulation` function in the report.
  """
  q=points(np.ones(d))
  qref=np.vstack([np.zeros((1,d)),np.tril(np.ones((d,d)))]).T
  P=perms(np.arange(d))
  nme=factorial(d);nq=2**d
  me=np.zeros((d+1,nme),dtype=int)
  A=np.matrix(2**np.arange(d))
  for k in range(nme):
    me[:,k]=A*qref[P[k]]
    ql=q[:,me[:,k]]
    if np.linalg.det(np.r_[np.ones((1,d+1)),ql])<0:
      tmp=me[0,k];me[0,k]=me[d,k];me[d,k]=tmp
  return q,me

def Triangulation(N):
  """ q,me=Triangulation(N)
  
      Compute the vertices array ``q`` and the connectivity array ``me`` obtained 
      from a tesselation of the cartesian grid CG(N) with simplices by using the Kuhn
      Triangulation.
      
      Parameters
      ----------
      N : list or array of d integers. The cartesian grid CG(N) is then defined on
             [0,N[0]]x...x[O,N[d-1]]
      
      Returns
      -------
      q : vertices array of the triangulation of CG(N), (d,nq)-shaped ndarray of 
          integers where ``nq=prod(N+1)`` and ``d=len(N)``
      me: connectivity array of the triangulation of CG(N), (d+1,nme) shaped ndarray 
          of integers  where ``nme=factorial(d)*prod(N). 
      
      As an example: 
      
      >>> q,me=Triangulation([20,100,30])
      >>> q
      array([[   0.,    1.,    2., ...,   18.,   19.,   20.],
             [   0.,    0.,    0., ...,  100.,  100.,  100.],
             [   0.,    0.,    0., ...,   30.,   30.,   30.]])
      >>> me
      array([[    0,  2143,  2143, ..., 63607, 63607, 65750],
             [    1,     1,    21, ..., 65728, 63628, 65728],
             [   22,  2122,    22, ..., 65729, 65749, 65749],
             [ 2143,     0,     0, ..., 65750, 65750, 63607]])
      
      See function `CGTriangulation` in the report.
  """
  N=np.array(N)
  d=len(N)
  q=points(N)
  Hinv=points(N-1)
  Nh=Hinv.shape[1]
  beta=betafunc(N)
  ibase=beta.dot(Hinv)
  qK,meK=KuhnTriangulation(d)
  fd=factorial(d)
  me=np.zeros((d+1,fd*Nh),dtype=int)
  Idx=fd*np.arange(Nh)
  for j in range(d+1):
    for l in range(fd):
      me[j,Idx+l]=ibase+beta.dot(qK[:,meK[j,l]])
  return q,me
    
def genTessFaces(N,m,**kwargs):
  """ Computes all m-faces tessellations of the cartesian grid CG(N) 
      with m-simplices or m-orthotopes.
      
      See `CGTriFaces` or `CGTessFaces` functions in the report.
  """
  from .EltMesh import EltMesh
  EltType=kwargs.get('type','simplex')
  assert(EltType in ['simplex','orthotope'])
  if EltType=='simplex':
    TessFun=lambda N: Triangulation(N)
  else:
    TessFun=lambda N: TessHyp(N)
  N=np.array(N)
  d=len(N)
  level=d-m
  beta=betafunc(N)
  sTh=[]
  if m==0:
    Q=np.diag(N).dot(points(np.ones(d)))
    ind=beta.dot(Q)
    for k in range(Q.shape[1]):
      ##sTh.append(smallTh(np.array([Q[:,k]]).T,np.array([[0]]),[int(ind[k])]))
      #sTh.append(smallTh(np.array([Q[:,k]]).T,np.array([[0]]),ind[k]))
      sTh.append(EltMesh(d,m,np.array([Q[:,k]]).T,np.array([[0]]),toGlobal=ind[k]))
    return sTh
  S=points(np.ones(level))
  #L=comb(d,d-m)
  L=combs(np.arange(d),d-m)
  nc=L.shape[0]
  # R=np.flipud(comb(d,m))
  R=np.flipud(combs(np.arange(d),m))
  for l in range(nc):
    qw,mew=TessFun(N[R[l,:]])
    nq=qw.shape[1]
    for r in range(2**level):
      q=np.zeros((d,nq))
      q[R[l],:]=qw
      tmp=np.matrix(N[L[l,:]].T*S[:,r]).T
      q[L[l],:]=tmp*np.ones(nq)
      #sTh.append(smallTh(q,mew,beta.dot(q)))
      sTh.append(EltMesh(d,m,q,mew,toGlobal=beta.dot(q)))
  return sTh
    
def TriFaces(N,m):    # TriFaces -> TessSimFaces
  """ sTh=TriFaces(N,m)
  
      Compute all m-faces tessellations of the cartesian grid CG(N) 
      with m-simplices.
      
      Parameters
      ----------
      N : list or array of d integers. The cartesian grid CG(N) is then defined on
             [0,N[0]]x...x[O,N[d-1]]
      m : integer (0<=m<=len(N)).
      
      Returns
      -------
      sTh :  list of EltMesh. Its length is the number of m-faces of CG(N):
          len(sTh)== 2**(d-m)*factorial(d)/(factorial(m)*factorial(d-m))
          
      :example:
      
      >>> sTh=TriFaces([20,100,30],2)
      >>> len(sTh)
      6
      >>> sTh[3]
      EltMesh object 
          type (str): simplicial
          type : 0
            d : 3
            m : 2
            q : (3,651)
            me : (3,1200)      
      
      See `CGTriFaces` function in the report.
  """
  return genTessFaces(N,m,type='simplex')

def TessFaces(N,m):     # TessFaces -> TessHypFaces
  """ sTh=TessFaces(N,m)
  
      Compute all m-faces tessellations of the cartesian grid CG(N) 
      with unit m-hypercubes.
      
            Parameters
      ----------
      N : list or array of d integers. The cartesian grid CG(N) is then defined on
             [0,N[0]]x...x[O,N[d-1]]
      m : integer (0<=m<=len(N)).
      
      Returns
      -------
      sTh :  list of EltMesh. Its length is the number of m-faces of CG(N):
          len(sTh)== 2**(d-m)*factorial(d)/(factorial(m)*factorial(d-m))
          
      :example:
      
      >>> sTh=TessFaces([20,100,30],2)
      >>> len(sTh)
      6
      >>> sTh[3]
      EltMesh object 
          type (str): orthotope
          type : 1
            d : 3
            m : 2
            q : (3,651)
            me : (4,600)
      
      See `CGTessFaces` function in the report.
  """
  return genTessFaces(N,m,type='orthotope')



""" ==============
    class OrthMesh
    ==============
    
    The aim of the class `OrthMesh` is to use functions of the `CartesianGrid` module  for 
    creating regular tessallations with simplices or orthotopes of an d-orthope and all 
    its m-faces (0<=m<=d).
"""

import numpy as np
from fc_tools.others import isModuleFound,LabelBaseName

class OrthMesh:
  """ 
    ==============
    class OrthMesh
    ==============
    
    The aim of the class `OrthMesh` is to use functions of the `CartesianGrid` 
    module  for creating regular tessallations with simplices or orthotopes of 
    an d-orthope and all its m-faces (0<=m<=d).
    The d-orthope to mesh is given in R^d by
      Oh=[a_1,b_1]x...x[a_d,b_d]
  
    Class attributes (main)
    -----------------------
    d     space dimension (integer).
    N     number of discretization in each space dimension or list of d 
          integers: one value by space dimension.
    type  type of the mesh elements, 'simplicial'(default) or 'orthotope' 
          (string).
    box   box corresponding to the d-orthotope to mesh. Could be given by a 
          2D list or array such that 
             a_i=box[i-1][0] and b_i=box[i-1][1]
          or
             a_i=box[i-1,0] and b_i=box[i-1,1]
    Mesh  main mesh as an `EltMesh` object, ``Mesh.q`` is the vertices array 
          and ``Mesh.me`` the connectivity array.
    Faces list of list of EltMesh objects. Faces[i] contains all meshes representing
          all  (d-i)-faces of the d-orthotope Oh. The k-th (d-i)-face is given
          by the `EltMesh` object ``Faces[i][k]``: its vertices array is 
          ``Faces[i][k].q`` and its the connectivity array ``Faces[i][k].me``.
          By construction, we have
             Mesh.q[:,Faces[i][k].toGlobal] == Faces[i][k].q
          
    Class properties
    ----------------
    constructor: ``Th=OrthMesh(d,N,**kwargs)``
          
  """
  
  def __init__(self,d,N,**kwargs):
    """ Th=OrthMesh(d,N,**kwargs)
        
        Parameters
        ----------
        d     space dimension 
        N     either a 1D array/list of length `d` such that `N[i-1]` is the number 
              of discretization for `[a_i,b_i]` or either an integer if the number 
              of discretization is the same in all space directions.
        
        Optional key/value parameters
        -----------------------------
        box=[[a_1,b_1],[a_2,b_2],...,[a_d,b_d]]
            Use to set the d-orthotope. By default, the unit reference d-hypercube
            is choosed: [0,1]^d
        type=value
            If value=='orthotope' then the tessallation is done with orthotopes.
            If value=='simplicial' then the tessallation is done with simplices.
        mapping=mapfun
            mapfun is a mapping function.
            
        :example:
        
        >>> Th=OrthMesh(3,[5,3,4])
        >>> Th.Mesh
        EltMesh object 
            type (str): simplicial
            type : 0
              d : 3
              m : 3
              q : (3,120)
              me : (4,360)
        >>> Th.Faces[1][0]
        EltMesh object 
            type (str): simplicial
            type : 0
              d : 3
              m : 1
              q : (3,5)
              me : (2,4)
        >>> Th.Mesh.q[:,Th.Faces[1][0].toGlobal] == Th.Faces[1][0].q
        array([[ True,  True,  True,  True,  True],
               [ True,  True,  True,  True,  True],
               [ True,  True,  True,  True,  True]], dtype=bool)
        
    """
    import fc_hypermesh.CartesianGrid as CG
    from fc_hypermesh.EltMesh import EltMesh
    from fc_tools.colors import selectColors

    ctype=kwargs.get('type', 'simplicial' )
    box=kwargs.get('box',np.ones((d,1))*np.array([0,1]))
    m_min=kwargs.get('m_min',0)
    mapping=kwargs.get('mapping',None)

    N=np.atleast_1d(N)
    assert len(N)==1 or len(N)==d
    if len(N)==1:
      N=N[0]*np.ones(d)
    if (ctype=='orthotope'):
      funMesh = lambda N: CG.TessHyp(N)
      funFaces = lambda N,m: CG.TessFaces(N,m)
      ntype=1
    else: # default 'simplicial'
      funMesh = lambda N: CG.Triangulation(N)
      funFaces = lambda N,m: CG.TriFaces(N,m)
      ntype=0
    self.d=d
    self.box=np.array(box,dtype=float)
    assert self.box.shape==(d,2)
    self.type=ctype
    [q,me]=funMesh(N)
    trans=lambda Q: MappingBox(Q,N,box)
    if mapping is not None:
      my_mapp=lambda Q: mapping(trans(Q))
    else:
      my_mapp=trans
    self.Mesh=EltMesh(d,d,my_mapp(q),me,label=1,type=ntype)
    self.set_box()
    self.Faces=[]
    i=0
    for m in np.arange(d-1,m_min-1,-1):
      self.Faces.append(funFaces(N,m))
      nsTh=len(self.Faces[i])
      colors = selectColors(nsTh)
      for j in np.arange(nsTh):
        self.Faces[i][j].q=my_mapp(self.Faces[i][j].q)
        self.Faces[i][j].label=j+1
        self.Faces[i][j].color=colors[j]
        self.Faces[i][j].type=ntype
      i+=1
      
  def __repr__(self):
    strret = ' %s object \n'%self.__class__.__name__ 
    strret += '      d : %d\n'%self.d 
    strret += '  Mesh  : %s\n'%str(self.Mesh)
    for F in self.Faces:
      strret += ' Number of %d-faces : %d\n'%(F[0].m,len(F))
      i=0
      for f in F:
        strret += '   [%2d] (type,nq,nme) : (%s,%d,%d)\n'%(i,f.strtype(),f.nq,f.nme)
        i+=1
    return strret         
   
  def getFacesIndex(self,m): # must be improved
    A=np.array(np.arange(self.d-1,-1,-1))
    return np.where(A==m)[0][0]
  
  def set_box(self):
    for i in range(self.d):
      self.box[i,0]=np.min(self.Mesh.q[i])
      self.box[i,1]=np.max(self.Mesh.q[i])
   
  def plotmesh(self,**kwargs):
    """ plotmesh(self,**kwargs)
    
    """
    if not isModuleFound('matplotlib'):
      print('plotmesh needs matplotlib package!')
      return
    import matplotlib.pyplot as plt
    if self.d > 3:
      print('Unable to plot in dimension %d > 3!'%self.d)
      return
    m=kwargs.pop('m',self.d)
    assert m in range(self.d+1)
    legend=kwargs.pop('legend', False);
    fig=plt.gcf()
    if m==self.d:
      Legend_handle,Labels=self.Mesh.plotmesh(**kwargs)
      Legend_handle=[Legend_handle]
      Labels=[Labels]
    else:
      Legend_handle=[];Labels=[]
      idx=self.getFacesIndex(m)
      F=self.Faces[idx]
      nF=len(F);
      for i in range(nF):
        Leg,Lab=F[i].plotmesh(**kwargs)
        Legend_handle.append(Leg)
        Labels.append(Lab)
    fig = plt.gcf()
    ax=fig.axes[0]
    if self.d==3:
      ax.set_zlim(self.box[2,0],self.box[2,1])
    ax.set_xlim(self.box[0,0],self.box[0,1])
    ax.set_ylim(self.box[1,0],self.box[1,1]) 
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    if self.d==3:
      ax.set_zlabel('z')
    #-> No background
    ax.patch.set_facecolor('None')
    fig.patch.set_visible(False) 
    #<-
    if legend:
      plt.legend(Legend_handle,Labels,loc='best', ncol=int(len(Legend_handle)/10)+1).draggable()
      
def MappingBox(q,N,box):
  box=np.array(box,dtype=float)
  d=len(N)
  
  for i in range(d):
    q[i]=box[i,0]+(box[i,1]-box[i,0])/(1.*N[i])*q[i] #BUG: take care with whole division
  return q


""" ====================
    fc_hypermesh package
    ====================
    
    The fc_hypermesh package contains a simple class object `OrthMesh` which 
    permits, in any dimension d>=1 to obtain a simplicial mesh or orthotope mesh 
    with all their m-faces, 0<=m<d,$ of a d-orthotope. If the `Matplotlib` package
    is installed, it is also possible with the `plotmesh` method of the class 
    object `OrthMesh` to represent a mesh or its m-faces for d<=3.
    
    ==================== =========================================================
    Objects
    ==============================================================================
    OrthMesh             Contain a main mesh of a d-orthotope and all its m-faces. 
                         See `OrthMesh` help.
    EltMesh              Elementary mesh (low level class).
                         See `EltMesh` help.
                         
    ==================== =========================================================
    Demo functions
    ==============================================================================
    See `demos` help
    
    ==================== =========================================================
    Benchmark functions
    ==============================================================================
    See `benchs` help
    
    Algorithms used in `fc_hypermesh` package are described in the report 
      'Vectorized algorithms for regular tessellations of d-orthotopes and 
       their faces' 
    by F. Cuvelier and available at http:\\...
    
    :author:     F. Cuvelier
    :email:      cuvelier@math.univ-paris13.fr
    :copyright:  (c) 2017
    :license:    GNU General Public License.
"""
__version__ = '0.0.12'
from fc_hypermesh.OrthMesh import OrthMesh
from fc_hypermesh.benchs import bench01
from fc_hypermesh import demos




from setuptools import setup, find_packages
from os import path
import sys
here = path.abspath(path.dirname(__file__))


if sys.version_info[0]>2:
  with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()
else:    
  with open(path.join(here, 'README.rst')) as f:  
    long_description = f.read().decode("UTF-8")


setup_defaults = {  
   'name'        : 'fc_hypermesh',
   'description' : 'Generate conforming meshes of hypercubes, hyperrectangles or of any d-orthotopes by simplices or orthotopes with their m-faces',
   'version'     : '0.0.12',
   'url'         : 'http://www.math.univ-paris13.fr/~cuvelier/software',
   'author'      : 'Francois Cuvelier',
   'author_email': 'cuvelier@math.univ-paris13.fr',
   'license'     : 'BSD',
   'packages'    : ['fc_hypermesh'],
   'classifiers':['Topic :: Scientific/Engineering :: Mathematics'],
   } 

setup(name=setup_defaults['name'],
      description = setup_defaults['description'],
      long_description=long_description,
      version=setup_defaults['version'],
      url=setup_defaults['url'],
      author=setup_defaults['author'],
      author_email=setup_defaults['author_email'],
      license = setup_defaults['license'],
      packages=setup_defaults['packages'],
      classifiers=setup_defaults['classifiers'],
      install_requires=['fc_tools >= 0.0.15']
     )